# ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: #
# Date: 07.09.2016                                                            #
# Author: Ole-Johan Skrede                                                    #
#                                                                             #
# Solution proposal as part of the exercise program in                        #
# INF4300 - Digital image analysis at the University of Oslo                  #
#                                                                             #
# ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: #
"""
Solution proposal for exercise 3, task 1.
"""

# pylint: disable=expression-not-assigned
# pylint: disable=bad-indentation
# pylint: disable=redefined-outer-name

import os
import argparse

import numpy as np
import cv2
import matplotlib.pyplot as plt

def line_detection_regular(image_filename, result_dir, line_threshold):
  """
  Detect lines in an image using regular Hough transform with polar coordinates.

  Args:
    image_filename: string. Full filepath for image to be analyzed.
    line_threshold: int. Number of points that has to be detected before it is
                    considered a line.
    min_hyst_thresh: int.
  """
  # Read 2D graylevel image
  image = cv2.imread(image_filename, cv2.IMREAD_COLOR)
  image_name = os.path.splitext(image_filename.split('/')[-1])[0]

  # Construct a binary image highlighting the edges in the image
  gray_image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
  gray_image = cv2.medianBlur(gray_image, 5)
  edges = cv2.Canny(gray_image,
                    100, # first threshold in hysteresis
                    200, # second threshold in hysteresis
                    apertureSize=3)

  # Detect straight lines using Hough transform
  lines = cv2.HoughLines(edges,
                         rho=1, # Distance resolution of the accumulator [pixels]
                         theta=np.pi/180, # Angle resolution [radians]
                         threshold=line_threshold)

  # Draw red lines on top of the original image
  lined_image = np.copy(image)
  line_length = 1000
  num_lines = 0
  for rho, theta in lines[:, 0]:
    a = np.cos(theta)
    b = np.sin(theta)
    x0 = a*rho
    y0 = b*rho
    x1 = int(x0 + line_length*(-b))
    y1 = int(y0 + line_length*(a))
    x2 = int(x0 - line_length*(-b))
    y2 = int(y0 - line_length*(a))

    cv2.line(lined_image, (x1, y1), (x2, y2), (0, 0, 255), 2)

    num_lines += 1

  print('Regular approach')
  print('Found {} lines'.format(num_lines))

  # Plot the results
  plt.figure(0)
  plt.imshow(image, cmap='gray')
  plt.title(image_name), plt.xticks([]), plt.yticks([])
  cv2.imwrite(os.path.join(result_dir, image_name + '_original.png'), image)

  plt.figure(1)
  plt.imshow(edges, cmap='gray')
  plt.title('Edges in '+image_name), plt.xticks([]), plt.yticks([])
  cv2.imwrite(os.path.join(result_dir, image_name + '_pp_regular.png'), edges)

  plt.figure(2)
  plt.imshow(cv2.cvtColor(lined_image, cv2.COLOR_BGR2RGB))
  plt.title('Lines in '+image_name), plt.xticks([]), plt.yticks([])
  cv2.imwrite(os.path.join(result_dir, image_name + '_lines_regular.png'), lined_image)

  # Illustration of rho-theta space
  ill_edges = cv2.Canny(gray_image, 100, 200, apertureSize=3)
  ill_lines = cv2.HoughLines(ill_edges, rho=1, theta=np.pi/180, threshold=0)
  ill_rho = ill_lines[:, 0, 0]
  ill_theta = ill_lines[:, 0, 1]

  heatmap, xedges, yedges = np.histogram2d(ill_theta, ill_rho, bins=200)
  extent = [xedges[0], xedges[-1], yedges[0], yedges[-1]]

  plt.figure(3)
  plt.imshow(heatmap, cmap='cubehelix', extent=extent, aspect='auto')
  plt.show()


def line_detection_probabilistic(image_filename, result_dir, min_line_length,
                                 max_line_gap, line_threshold):
  """
  Detect lines in an image using regular Hough transform with polar coordinates.

  Args:
    image_filename: string. Full filepath for image to be analyzed.
    min_line_length: int. Minimum length of line, shorter segments will be
                     rejected.
    max_line_gap: int. Maximum allowed gap between line segments to treat them
                  as single lines.
    line_threshold: int. Number of points that has to be detected before it is
                    considered a line.
  """
  # Read 2D graylevel image
  image = cv2.imread(image_filename, cv2.IMREAD_COLOR)
  image_name = os.path.splitext(image_filename.split('/')[-1])[0]

  # Construct a binary image highlighting the edges in the image
  gray_image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
  gray_image = cv2.medianBlur(gray_image, 5)
  edges = cv2.Canny(gray_image,
                    100, # First hysteresis threshold
                    150, # Second hysteresis threshold,
                    apertureSize=3)

  # Detect straight lines using Hough transform
  lines = cv2.HoughLinesP(edges,
                          rho=1, # Distance resolution
                          theta=np.pi/180, # Angle resolution
                          threshold=line_threshold,
                          minLineLength=min_line_length,
                          maxLineGap=max_line_gap)

  # Draw red lines on top of the original image
  lined_image = np.copy(image)
  num_lines = 0
  for x1, y1, x2, y2 in lines[:, 0]:
    cv2.line(lined_image, (x1, y1), (x2, y2), (0, 0, 255), 2)

    num_lines += 1

  print('Probabilistic approach')
  print('Found {} lines'.format(num_lines))

  # Plot the results
  plt.figure(0)
  plt.imshow(image, cmap='gray')
  plt.title(image_name), plt.xticks([]), plt.yticks([])
  cv2.imwrite(os.path.join(result_dir, image_name + '_original.png'), image)

  plt.figure(1)
  plt.imshow(edges, cmap='gray')
  plt.title('Edges in '+image_name), plt.xticks([]), plt.yticks([])
  cv2.imwrite(os.path.join(result_dir, image_name + '_pp_prob.png'), edges)

  plt.figure(2)
  plt.imshow(cv2.cvtColor(lined_image, cv2.COLOR_BGR2RGB))
  plt.title('Lines in '+image_name), plt.xticks([]), plt.yticks([])
  cv2.imwrite(os.path.join(result_dir, image_name + '_lines_prob.png'), lined_image)

def circle_detection(image_filename, result_dir, min_dist,
                     canny_thresh, acc_thresh):
  """
  Detect lines in an image using regular Hough transform with polar coordinates.

  Args:
    image_filename: string. Full filepath for image to be analyzed.
    canny_thresh: int. Highest threshold in the hysteresis used in Canny edge
                  detection, the lowest is set to canny_thresh/2
    acc_thresh: int. The accumulator threshold for the circle centers at the
                detections stage. Smaller can mean more false positive detected
                circles.
  """
  # Read 2D graylevel image
  image = cv2.imread(image_filename, cv2.IMREAD_COLOR)
  image_name = os.path.splitext(image_filename.split('/')[-1])[0]

  # Convert to graylevels and blur
  pp_image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
  pp_image = cv2.medianBlur(pp_image, 9)

  # Detect straight lines using Hough transform (note that this has a built-in
  # edge detector (Cannys algorithm)).
  circles = cv2.HoughCircles(pp_image,
                             method=cv2.HOUGH_GRADIENT,
                             dp=1,
                             minDist=min_dist,
                             param1=canny_thresh,
                             param2=acc_thresh,
                             minRadius=0,
                             maxRadius=0)
  circles = np.uint16(np.around(circles))

  # Draw red lines on top of the original image
  circled_image = np.copy(image)
  num_circles = 0
  for x, y, r in circles[0, :]:
    # Draw outer circle
    cv2.circle(circled_image, (x, y), r, (0, 255, 0), 2)

    # Draw center point
    cv2.circle(circled_image, (x, y), 2, (0, 0, 255), 3)

    num_circles += 1

  print('Circle detection')
  print('Found {} circles'.format(num_circles))

  # Plot the results
  plt.figure(0)
  plt.imshow(image, cmap='gray')
  plt.title(image_name), plt.xticks([]), plt.yticks([])
  cv2.imwrite(os.path.join(result_dir, image_name + '_original.png'), image)

  plt.figure(1)
  plt.imshow(pp_image, cmap='gray')
  plt.title('Blurred '+image_name), plt.xticks([]), plt.yticks([])
  cv2.imwrite(os.path.join(result_dir, image_name + '_pp.png'), pp_image)

  plt.figure(2)
  plt.imshow(cv2.cvtColor(circled_image, cv2.COLOR_BGR2RGB))
  plt.title('Circles in '+image_name), plt.xticks([]), plt.yticks([])
  cv2.imwrite(os.path.join(result_dir, image_name + '_circles.png'), circled_image)



def main():
  """Line detection implementations in openCV"""

  parser = argparse.ArgumentParser()
  parser.add_argument('-t', '--task', default='line_regular',
                      help='{line_regular, line_probabilistic, circle}')
  parser.add_argument('-i', '--in_filepath', required=True,
                      help='Full filepath of input image (grayscale image)')
  parser.add_argument('-o', '--result_dir', default='./results/task01',
                      help='Output filepath (is created if it not exists)')
  args = parser.parse_args()

  print('='*80)
  print('Solution to weekly exercises in INF4300')
  print('Exercise 3')
  print('Task 1')
  print('-'*80)

  # Configurations

  if not os.path.exists(args.result_dir):
    print("Creating result folder: {}".format(args.result_dir))
    os.makedirs(args.result_dir)

  if args.task == 'line_regular':
    line_threshold_reg = 150

    line_detection_regular(args.in_filepath, args.result_dir, line_threshold_reg)

  if args.task == 'line_probabilistic':
    line_threshold_prob = 200
    min_line_length = 40
    max_line_gap = 30

    line_detection_probabilistic(args.in_filepath, args.result_dir, line_threshold_prob,
                                 min_line_length, max_line_gap)

  if args.task == 'circle':
    min_dist = 30
    canny_thresh = 150
    acc_thresh = 30

    circle_detection(args.in_filepath, args.result_dir, min_dist, canny_thresh, acc_thresh)

  plt.show()

if __name__ == "__main__":
  main()
